@@ -2,7 +2,7 @@
 This file is part of bruteforce-salted-openssl, a program trying to
 bruteforce a file encrypted (with salt) by openssl.
 
-Copyright 2014-2018 Guillaume LE VAILLANT
+Copyright 2014-2024 Guillaume LE VAILLANT
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -49,6 +49,7 @@ the covered work.
 
 
 #define LAST_PASS_MAX_SHOWN_LENGTH 256
+#define EXIT_NO_PASSWORD_FOUND 2
 
 unsigned char *default_charset = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
 
@@ -78,13 +79,12 @@ FILE *dictionary = NULL;
 const EVP_CIPHER *cipher = NULL;
 const EVP_MD *digest = NULL;
 pthread_mutex_t found_password_lock, get_password_lock;
-char stop = 0, only_one_password = 0, found_password = 0, no_error = 0, no_salt = 0;
+char stop = 0, only_one_password = 0, found_password = 0, no_error = 0, no_salt = 0, pbkdf2 = 0, debug = 0;
 unsigned int nb_threads = 1;
-unsigned long long int limit = 0, count_limit = 0;
+unsigned long long int limit = 0, count_limit = 0, iterations = 10000;
 unsigned char last_pass[LAST_PASS_MAX_SHOWN_LENGTH];
 time_t start_time;
 unsigned int status_interval = 0;
-struct itimerval progress_timer, state_timer;
 struct decryption_func_locals
 {
   unsigned long long int counter;
@@ -97,7 +97,7 @@ unsigned int *tab = NULL;
  * Statistics
  */
 
-void handle_signal(int signo)
+void print_status()
 {
   unsigned long long int total_ops = 0;
   unsigned int i, l;
@@ -130,7 +130,11 @@ void handle_signal(int signo)
     fprintf(stderr, "Tried passwords: %llu\n", total_ops);
   fprintf(stderr, "Tried passwords per second: %lf\n", pw_per_seconds);
   if(binary == NULL)
+  {
+    pthread_mutex_lock(&get_password_lock);
     fprintf(stderr, "Last tried password: %s\n", last_pass);
+    pthread_mutex_unlock(&get_password_lock);
+  }
   if(dictionary == NULL)
   {
     fprintf(stderr, "Total space searched: %lf%%\n", (total_ops / space) * 100);
@@ -150,6 +154,31 @@ void handle_signal(int signo)
   fprintf(stderr, "\n");
 }
 
+void * status_func(void *arg)
+{
+  unsigned int i;
+
+  while(stop == 0)
+  {
+    for(i = 0; i < status_interval; i++)
+    {
+      sleep(1);
+      if(stop == 1)
+      {
+        pthread_exit(NULL);
+      }
+    }
+    print_status();
+  }
+
+  pthread_exit(NULL);
+}
+
+void handle_signal(int signo)
+{
+  print_status();
+}
+
 
 /*
  * Decryption
@@ -384,6 +413,18 @@ int read_dictionary_line(unsigned char **line, unsigned int *n)
   return(1);
 }
 
+void print_hex(unsigned char *buf, int len)
+{
+  int i;
+  int n;
+
+  for(i = 0, n = 0; i < len; i++)
+  {
+    printf("%02x",buf[i]);
+  }
+  printf("\n");
+}
+
 void * decryption_func(void *arg)
 {
   struct decryption_func_locals *dfargs;
@@ -394,8 +435,14 @@ void * decryption_func(void *arg)
 
   dfargs = (struct decryption_func_locals *) arg;
 
-  key = (unsigned char *) malloc(EVP_CIPHER_key_length(cipher));
-  iv = (unsigned char *) malloc(EVP_CIPHER_iv_length(cipher));
+  int iklen = EVP_CIPHER_key_length(cipher);
+  int ivlen = EVP_CIPHER_iv_length(cipher);
+
+  key = (unsigned char *) malloc(iklen);
+  iv = (unsigned char *) malloc(ivlen);
+
+  char *keyivpair = (unsigned char *) malloc(iklen + ivlen);
+
   out = (unsigned char *) malloc(data_len + EVP_CIPHER_block_size(cipher));
   ctx = EVP_CIPHER_CTX_new();
   if((key == NULL) || (iv == NULL) || (out == NULL) || (ctx == NULL))
@@ -420,9 +467,42 @@ void * decryption_func(void *arg)
 
     /* Decrypt data with password */
     if(no_salt)
-      EVP_BytesToKey(cipher, digest, NULL, pwd, pwd_len, 1, key, iv);
+    {
+      if(pbkdf2)
+      {
+        PKCS5_PBKDF2_HMAC(pwd, pwd_len, NULL, 0, iterations, digest, iklen + ivlen, keyivpair);
+        memcpy(key, keyivpair, iklen);
+        memcpy(iv, keyivpair + iklen, ivlen);
+      }
+      else
+      {
+        EVP_BytesToKey(cipher, digest, NULL, pwd, pwd_len, 1, key, iv);
+      }
+    }
     else
-      EVP_BytesToKey(cipher, digest, salt, pwd, pwd_len, 1, key, iv);
+    {
+      if(pbkdf2)
+      {
+        PKCS5_PBKDF2_HMAC(pwd, pwd_len, salt, sizeof(salt), iterations, digest, iklen + ivlen, keyivpair);
+        memcpy(key, keyivpair, iklen);
+        memcpy(iv, keyivpair + iklen, ivlen);
+      }
+      else
+      {
+        EVP_BytesToKey(cipher, digest, salt, pwd, pwd_len, 1, key, iv);
+      }
+    }
+
+    if(debug)
+    {
+      printf("pwd: %s\n", pwd);
+      printf("salt: ");
+      print_hex(salt, sizeof(salt));
+      printf("key: ");
+      print_hex(key, 32);
+      printf("iv: ");
+      print_hex(iv, 16);
+    }
 
     EVP_DecryptInit(ctx, cipher, key, iv);
 
@@ -447,11 +527,15 @@ void * decryption_func(void *arg)
     }
 
     /* Don't bother checking the rest if the first preview part didn't match. */
-    if(preview_found) {
-      EVP_DecryptUpdate(ctx, out + total_out_len, &cur_out_len, data + preview_len, data_len - preview_len);
-      total_out_len += cur_out_len;
-      ret = EVP_DecryptFinal(ctx, out + total_out_len, &cur_out_len);
-      total_out_len += cur_out_len;
+    if(preview_found)
+    {
+      ret = EVP_DecryptUpdate(ctx, out + total_out_len, &cur_out_len, data + preview_len, data_len - preview_len);
+      if(ret == 1)
+      {
+        total_out_len += cur_out_len;
+        ret = EVP_DecryptFinal(ctx, out + total_out_len, &cur_out_len);
+        total_out_len += cur_out_len;
+      }
     }
 
     if(no_error || (ret == 1))
@@ -467,7 +551,7 @@ void * decryption_func(void *arg)
     if(found)
     {
       /* We have a positive result */
-      handle_signal(SIGUSR1); /* Print some stats */
+      print_status();
       pthread_mutex_lock(&found_password_lock);
       found_password++;
       printf("Password candidate: %s\n", pwd);
@@ -508,7 +592,7 @@ void * decryption_func(void *arg)
  * Save/restore state
  */
 
-void save_state(int signo)
+void save_state()
 {
   unsigned int i;
   unsigned long long int total_ops = 0;
@@ -552,6 +636,26 @@ void save_state(int signo)
   fclose(state);
 }
 
+void * state_func(void *arg)
+{
+  unsigned int i;
+
+  while(stop == 0)
+  {
+    for(i = 0; i < 60; i++)
+    {
+      sleep(1);
+      if(stop == 1)
+      {
+        pthread_exit(NULL);
+      }
+    }
+    save_state();
+  }
+
+  pthread_exit(NULL);
+}
+
 void restore_state()
 {
   unsigned int i, n;
@@ -787,12 +891,16 @@ void usage(char *progname)
   fprintf(stderr, "                 default: \"\"\n\n");
   fprintf(stderr, "  -c <cipher>  Cipher for decryption.\n");
   fprintf(stderr, "                 default: aes-256-cbc\n\n");
+  fprintf(stderr, "  -D           Display salt, key and iv along with the tested password.\n\n");
   fprintf(stderr, "  -d <digest>  Digest for key and initialization vector generation.\n");
-  fprintf(stderr, "                 default: md5\n\n");
+  fprintf(stderr, "                 default: sha256\n\n");
   fprintf(stderr, "  -e <string>  End of the password.\n");
   fprintf(stderr, "                 default: \"\"\n\n");
   fprintf(stderr, "  -f <file>    Read the passwords from a file instead of generating them.\n\n");
   fprintf(stderr, "  -h           Show help and quit.\n\n");
+  fprintf(stderr, "  -i <n>       With -K, set the PBKDF2 iteration count to <n>.\n");
+  fprintf(stderr, "                 default: 10000\n\n");
+  fprintf(stderr, "  -K           Use PKCS5_PBKDF2_HMAC to derive the key.\n\n");
   fprintf(stderr, "  -L <n>       Limit the maximum number of tested passwords to <n>.\n\n");
   fprintf(stderr, "  -l <length>  Minimum password length (beginning and end included).\n");
   fprintf(stderr, "                 default: 1\n\n");
@@ -800,13 +908,13 @@ void usage(char *progname)
   fprintf(stderr, "               with <string>. Without this option, the decryption is considered\n");
   fprintf(stderr, "               as successful when the data contains mostly printable ASCII\n");
   fprintf(stderr, "               characters (at least 90%%).\n\n");
-  fprintf(stderr, "  -p <n>       Preview and check the first N decrypted bytes for the magic string.\n");
-  fprintf(stderr, "               If the magic string is present, try decrypting the rest of the data.\n");
-  fprintf(stderr, "                 default: 1024\n\n");
   fprintf(stderr, "  -m <length>  Maximum password length (beginning and end included).\n");
   fprintf(stderr, "                 default: 8\n\n");
   fprintf(stderr, "  -N           Ignore decryption errors (similar to openssl -nopad).\n\n");
   fprintf(stderr, "  -n           Ignore salt (similar to openssl -nosalt).\n\n");
+  fprintf(stderr, "  -p <n>       Preview and check the first N decrypted bytes for the magic string.\n");
+  fprintf(stderr, "               If the magic string is present, try decrypting the rest of the data.\n");
+  fprintf(stderr, "                 default: 1024\n\n");
   fprintf(stderr, "  -s <string>  Password character set.\n");
   fprintf(stderr, "               default: \"0123456789ABCDEFGHIJKLMNOPQRSTU\n");
   fprintf(stderr, "                         VWXYZabcdefghijklmnopqrstuvwxyz\"\n\n");
@@ -832,7 +940,7 @@ void list_algorithms(void)
 
 int main(int argc, char **argv)
 {
-  pthread_t *decryption_threads;
+  pthread_t *decryption_threads, status_thread, state_thread;
   char *filename;
   int fd, i, ret, c;
   struct stat file_stats;
@@ -842,7 +950,7 @@ int main(int argc, char **argv)
 
   /* Get options and parameters */
   opterr = 0;
-  while((c = getopt(argc, argv, "1aB:b:c:d:e:f:hL:l:M:m:Nns:t:v:w:p:")) != -1)
+  while((c = getopt(argc, argv, "1aB:b:c:d:De:f:hi:KL:l:M:m:Nns:t:v:w:p:")) != -1)
     switch(c)
     {
     case '1':
@@ -883,6 +991,10 @@ int main(int argc, char **argv)
       }
       break;
 
+    case 'D':
+      debug = 1;
+      break;
+
     case 'd':
       digest = EVP_get_digestbyname(optarg);
       if(digest == NULL)
@@ -922,6 +1034,14 @@ int main(int argc, char **argv)
       exit(EXIT_FAILURE);
       break;
 
+    case 'i':
+      iterations = (long unsigned int) atol(optarg);
+      break;
+
+    case 'K':
+      pbkdf2 = 1;
+      break;
+
     case 'L':
       limit = (long unsigned int) atol(optarg);
       break;
@@ -946,6 +1066,10 @@ int main(int argc, char **argv)
       no_salt = 1;
       break;
 
+    case 'p':
+      preview_len = (unsigned int) atoi(optarg);
+      break;
+
     case 's':
       charset_len = mbstowcs(NULL, optarg, 0);
       if(charset_len == 0)
@@ -981,10 +1105,6 @@ int main(int argc, char **argv)
       state_file = optarg;
       break;
 
-    case 'p':
-      preview_len = (unsigned int) atoi(optarg);
-      break;
-
     default:
       usage(argv[0]);
       switch(optopt)
@@ -999,11 +1119,11 @@ int main(int argc, char **argv)
       case 'l':
       case 'M':
       case 'm':
+      case 'p':
       case 's':
       case 't':
       case 'v':
       case 'w':
-      case 'p':
         fprintf(stderr, "Error: missing argument for option: '-%c'.\n\n", optopt);
         break;
 
@@ -1028,7 +1148,7 @@ int main(int argc, char **argv)
   if(cipher == NULL)
     cipher = EVP_aes_256_cbc();
   if(digest == NULL)
-    digest = EVP_md5();
+    digest = EVP_sha256();
   if(dictionary != NULL)
   {
     fprintf(stderr, "Warning: using dictionary mode, ignoring options -b, -e, -l, -m and -s.\n\n");
@@ -1125,6 +1245,7 @@ int main(int argc, char **argv)
       exit(EXIT_FAILURE);
     }
   }
+  path = filename;
 
   /* Read encrypted data */
   ret = fstat(fd, &file_stats);
@@ -1170,12 +1291,12 @@ int main(int argc, char **argv)
   signal(SIGUSR1, handle_signal);
   if(status_interval > 0)
   {
-    signal(SIGALRM, handle_signal);
-    progress_timer.it_value.tv_sec = status_interval;
-    progress_timer.it_value.tv_usec = 0;
-    progress_timer.it_interval.tv_sec = status_interval;
-    progress_timer.it_interval.tv_usec = 0;
-    setitimer(ITIMER_REAL, &progress_timer, NULL);
+    ret = pthread_create(&status_thread, NULL, &status_func, NULL);
+    if(ret != 0)
+    {
+      perror("Error: status thread");
+      exit(EXIT_FAILURE);
+    }
   }
 
   pthread_mutex_init(&found_password_lock, NULL);
@@ -1195,12 +1316,12 @@ int main(int argc, char **argv)
   {
     restore_state();
 
-    signal(SIGVTALRM, save_state);
-    state_timer.it_value.tv_sec = 60 * nb_threads;
-    state_timer.it_value.tv_usec = 0;
-    state_timer.it_interval.tv_sec = 60 * nb_threads;
-    state_timer.it_interval.tv_usec = 0;
-    setitimer(ITIMER_VIRTUAL, &state_timer, NULL);
+    ret = pthread_create(&state_thread, NULL, &state_func, NULL);
+    if(ret != 0)
+    {
+      perror("Error: state thread");
+      exit(EXIT_FAILURE);
+    }
   }
 
   /* Start decryption threads */
@@ -1218,9 +1339,18 @@ int main(int argc, char **argv)
   {
     pthread_join(decryption_threads[i], NULL);
   }
+  stop = 1;
+  if(state_file != NULL)
+  {
+    pthread_join(state_thread, NULL);
+  }
+  if(status_interval > 0)
+  {
+    pthread_join(status_thread, NULL);
+  }
   if(found_password == 0)
   {
-    handle_signal(SIGUSR1); /* Print some stats */
+    print_status();
     fprintf(stderr, "Password not found.\n");
     fprintf(stderr, "The file might have been encrypted with a different cipher or/and a\n");
     fprintf(stderr, "different digest (e.g. OpenSSL 1.0.x uses the MD5 digest by default\n");
@@ -1234,5 +1364,12 @@ int main(int argc, char **argv)
   free(data);
   EVP_cleanup();
 
-  exit(EXIT_SUCCESS);
+  if(found_password == 0)
+  {
+    return(EXIT_NO_PASSWORD_FOUND);
+  }
+  else
+  {
+    return(EXIT_SUCCESS);
+  }
 }
